import hvac
import os
import re


def resolve_secret(value, client):
    re_vault = re.compile("^VAULT:([^:]+):([^:]+)$")
    if not value.startswith("VAULT:"):
        return False, value
    m = re_vault.match(value)
    if not m:
        raise Exception("Invalid vault accessor '{}'".format(value))
    path, key = m.groups()
    data = client.read(path)
    if not data:
        raise Exception("Did not find secret at '{}'".format(path))

    if key not in data["data"]:
        raise Exception("Did not find key '{}' at secret path '{}'".format(
            key, path))
    return True, data["data"][key]


def resolve_secrets(x, client):
    if not x:
        pass
    elif type(x) == dict:
        resolve_secrets_dict(x, client)
    else:
        resolve_secrets_object(x, client)


def resolve_secrets_object(obj, client):
    for k, v in vars(obj).items():
        if type(v) == str:
            updated, v = resolve_secret(v, client)
            if updated:
                setattr(obj, k, v)


def resolve_secrets_dict(d, client):
    for k, v, in d.items():
        if type(v) == str:
            updated, v = resolve_secret(v, client)
            if updated:
                d[k] = v


class vault_config:
    def __init__(self, url, auth_method, auth_args):
        self.url = url
        self.auth_method = auth_method
        self.auth_args = auth_args

    def client(self):
        if not self.url:
            return vault_not_enabled()
        # NOTE: we might actually try and pick up VAULT_TOKEN from the
        # environment, but can't let that value override any value
        # passed here.
        # See for this workaround https://github.com/hvac/hvac/issues/421
        drop_envvar("VAULT_ADDR")
        drop_envvar("VAULT_TOKEN")

        if self.auth_method == "token":
            cl = hvac.Client(url=self.url, token=self.auth_args["token"])
        else:
            cl = hvac.Client(url=self.url)
            print("Authenticating with the vault using '{}'".format(
                self.auth_method))

            if self.auth_method == "github":
                if not self.auth_args:
                    self.auth_args = {}
                if "token" not in self.auth_args:
                    self.auth_args["token"] = get_github_token()

            getattr(cl.auth, self.auth_method).login(**self.auth_args)
        return cl


class vault_not_enabled:
    def __getattr__(self, name):
        raise Exception("Vault access is not enabled")


def get_github_token():
    try:
        return os.environ["VAULT_AUTH_GITHUB_TOKEN"]
    except KeyError:
        return input("Enter GitHub token for vault: ").strip()


def drop_envvar(name):
    if name in os.environ:
        del os.environ[name]



import os
import tempfile

import docker
from PIL import Image

from orderly_web.config import build_config
from orderly_web.notify import Notifier
from orderly_web.status import status
from orderly_web.pull import pull
from orderly_web.docker_helpers import docker_client, \
    ensure_network, ensure_volume, container_wait_running, \
    exec_safely, string_into_container, file_into_container


def start(path, extra=None, options=None, pull_images=False):
    st = status(path)
    for name, data in st.containers.items():
        if data["status"] is not "missing":
            msg = "Container '{}' is {}; please run orderly-web stop".format(
                name, data["status"])
            print(msg)
            return False
    cfg = build_config(path, extra, options)
    cfg.resolve_secrets()
    if pull_images:
        pull(cfg)

    notifier = Notifier(cfg.slack_webhook_url)
    with docker_client() as cl:
        notifier.post("*Starting* deploy to {}".format(cfg.web_url))
        try:
            init_all(cl, cfg)
            notifier.post("*Completed* deploy to {} :shipit:"
                          .format(cfg.web_url))
            return True
        except Exception:
            notifier.post("*Failed* deploy to {} :bomb:".format(cfg.web_url))
            raise


def init_all(cl, cfg):
    ensure_network(cl, cfg.network)
    for v in cfg.volumes.values():
        ensure_volume(cl, v)
    orderly_init(cfg, cl)
    web_init(cfg, cl)
    proxy_init(cfg, cl)
    config_save(cfg)


def orderly_init(cfg, docker_client):
    container = orderly_container(cfg, docker_client)
    orderly_write_ssh_keys(cfg.orderly_ssh, container)
    orderly_initial_data(cfg, container)
    orderly_check_schema(container)
    orderly_write_env(cfg.orderly_env, container)
    orderly_start(container)
    return container


def orderly_container(cfg, docker_client):
    print("Creating orderly container")
    args = ["--port", "8321", "--go-signal", "/go_signal", "/orderly"]
    image = str(cfg.images["orderly"])
    mounts = [docker.types.Mount("/orderly", cfg.volumes["orderly"])]
    container = docker_client.containers.run(
        image, args, mounts=mounts, network=cfg.network,
        name=cfg.containers["orderly"], working_dir="/orderly", detach=True)
    return container


def orderly_initial_data(cfg, container):
    if orderly_is_initialised(container):
        print("orderly volume already contains data - not initialising")
    else:
        if cfg.orderly_initial_source == "demo":
            orderly_init_demo(container)
        elif cfg.orderly_initial_source == "clone":
            orderly_init_clone(container, cfg.orderly_initial_url)
        else:
            raise Exception("Orderly volume not initialised")


def orderly_write_env(env, container):
    if not env:
        return
    print("Writing orderly environment")
    dest = "/root/.Renviron"
    txt = "".join(["{}={}\n".format(str(k), str(v)) for k, v in env.items()])
    string_into_container(txt, container, dest)


def orderly_init_demo(container):
    print("Initialising orderly with demo data")
    args = ["Rscript", "-e", "orderly:::create_orderly_demo('/orderly')"]
    exec_safely(container, args)


def orderly_init_clone(container, url):
    print("Initialising orderly by cloning")
    args = ["git", "clone", url, "/orderly"]
    exec_safely(container, args)


def orderly_is_initialised(container):
    res = container.exec_run(["stat", "/orderly/orderly_config.yml"])
    return res[0] == 0


def orderly_check_schema(container):
    print("Checking orderly schema is current")
    exec_safely(container, ["orderly", "rebuild", "--if-schema-changed"])


def orderly_write_ssh_keys(orderly_ssh, container):
    if not orderly_ssh:
        return
    print("Configuring ssh")
    path_private = "/root/.ssh/id_rsa"
    path_public = "/root/.ssh/id_rsa.pub"
    path_known_hosts = "/root/.ssh/known_hosts"
    exec_safely(container, ["mkdir", "-p", "/root/.ssh"])
    string_into_container(orderly_ssh["private"], container, path_private)
    string_into_container(orderly_ssh["public"], container, path_public)
    exec_safely(container, ["chmod", "600", path_private])
    hosts = exec_safely(container, ["ssh-keyscan", "github.com"])
    string_into_container(hosts[1].decode("UTF-8"), container,
                          path_known_hosts)


def orderly_start(container):
    print("Starting orderly server")
    exec_safely(container, ["touch", "/go_signal"])


def web_init(cfg, docker_client):
    if cfg.sass_variables is not None:
        web_generate_css(cfg, docker_client)
    container = web_container(cfg, docker_client)
    web_container_config(cfg, container)
    web_migrate(cfg, docker_client)
    web_start(container)
    return container


def web_container(cfg, docker_client):
    print("Creating web container")
    image = str(cfg.images["web"])
    mounts = [docker.types.Mount("/orderly", cfg.volumes["orderly"])]
    if cfg.sass_variables is not None:
        mounts.append(docker.types.Mount("/static/public/css",
                                         cfg.volumes["css"]))
    if cfg.logo_name is not None:
        logo_in_container = "/static/public/img/logo/{}".format(cfg.logo_name)
        mounts.append(docker.types.Mount(logo_in_container,
                                         cfg.logo_path,
                                         type="bind"))
    if cfg.web_dev_mode:
        port = cfg.web_port
        # NOTE: different format to proxy below because we only
        # expose this to the localhost, and not to any external
        # interface
        ports = {"{}/tcp".format(cfg.web_port): ("127.0.0.1", cfg.web_port)}
    else:
        ports = None
    container = docker_client.containers.run(
        image, mounts=mounts, network=cfg.network, ports=ports,
        name=cfg.containers["web"], detach=True)
    if cfg.favicon_path is not None:
        img = generate_favicon(cfg.favicon_path)
        file_into_container(img,
                            container,
                            "/static/public",
                            "favicon.ico")
        os.remove(img)
    return container


def generate_favicon(source):
    img = Image.open(source)
    fd, name = tempfile.mkstemp()
    img.save(name, format="ico")
    return name


def web_container_config(cfg, container):
    print("Configuring web container")
    orderly_container = cfg.containers["orderly"]
    opts = {"app.port": str(cfg.web_port),
            "app.name": cfg.web_name,
            "app.email": cfg.web_email,
            "app.url": cfg.web_url,
            "auth.github_org": cfg.web_auth_github_org or "",
            "auth.github_team": cfg.web_auth_github_team or "",
            "auth.fine_grained": str(cfg.web_auth_fine_grained).lower(),
            "auth.provider": "montagu" if cfg.web_auth_montagu else "github",
            "orderly.server": "http://{}:8321".format(orderly_container)}
    if cfg.logo_name is not None:
        opts["app.logo"] = cfg.logo_name
    if cfg.web_auth_montagu:
        opts["montagu.url"] = cfg.montagu_url
        opts["montagu.api_url"] = cfg.montagu_api_url
    if cfg.web_auth_github_app:
        opts["auth.github_key"] = cfg.web_auth_github_app["id"]
        opts["auth.github_secret"] = cfg.web_auth_github_app["secret"]

    txt = "".join(["{}={}\n".format(k, v) for k, v in opts.items()])
    exec_safely(container, ["mkdir", "-p", "/etc/orderly/web"])
    string_into_container(txt, container, "/etc/orderly/web/config.properties")


def web_migrate(cfg, docker_client):
    print("Migrating the web tables")
    image = str(cfg.images["migrate"])
    mounts = [docker.types.Mount("/orderly", cfg.volumes["orderly"])]
    docker_client.containers.run(image, mounts=mounts, remove=True)


def web_generate_css(cfg, docker_client):
    print("Generating custom css")
    image = str(cfg.images["css-generator"])
    compiled_css_mount = \
        docker.types.Mount("/static/public/css", cfg.volumes["css"])
    variable_mount = \
        docker.types.Mount("/static/src/scss/partials/user-variables.scss",
                           cfg.sass_variables, type="bind")

    mounts = [compiled_css_mount, variable_mount]
    docker_client.containers.run(image, mounts=mounts, remove=True)


def web_start(container):
    print("Starting orderly server")
    exec_safely(container, ["touch", "/etc/orderly/web/go_signal"])


def proxy_init(cfg, docker_client):
    if not cfg.proxy_enabled:
        return
    container = proxy_container(cfg, docker_client)
    proxy_certificates(cfg, container)


def proxy_container(cfg, docker_client):
    print("Creating proxy container")
    image = str(cfg.images["proxy"])
    orderly = "{}:{}".format(cfg.containers["web"], cfg.web_port)
    args = [cfg.proxy_hostname, str(cfg.proxy_port_http),
            str(cfg.proxy_port_https), orderly]
    mounts = [docker.types.Mount("/var/log/nginx", cfg.volumes["proxy_logs"])]
    ports = {
        "{}/tcp".format(cfg.proxy_port_http): cfg.proxy_port_http,
        "{}/tcp".format(cfg.proxy_port_https): cfg.proxy_port_https
    }
    ret = docker_client.containers.run(
        image, args, detach=True, name=cfg.containers["proxy"],
        network=cfg.network, mounts=mounts, ports=ports)
    return container_wait_running(ret)


def proxy_certificates(cfg, container):
    if cfg.proxy_ssl_self_signed:
        print("Generating self-signed certificates for proxy")
        exec_safely(container, ["self-signed-certificate", "/run/proxy"])
    else:
        print("Copying ssl certificate and key into proxy")
        string_into_container(cfg.proxy_ssl_certificate, container,
                              "/run/proxy/certificate.pem")
        string_into_container(cfg.proxy_ssl_key, container,
                              "/run/proxy/key.pem")


def config_save(cfg):
    print("Persisting configuration")
    cfg.save()



import docker

from orderly_web.config import fetch_config, build_config
from orderly_web.docker_helpers import *
from orderly_web.errors import OrderlyWebConfigError


def stop(path, kill=False, network=False, volumes=False, force=False,
         extra=None, options=None):
    try:
        cfg = fetch_config(path)
    except docker.errors.NotFound as e:
        if force:
            print("Unable to fetch config from orderly-web, forcing stop.")
            cfg = build_config(path, extra, options)
        else:
            msg = ("Unable to fetch config from orderly-web. To force stop, "
                   "provide --force option and any configuration options in "
                   "--extra and --options.")
            raise OrderlyWebConfigError(msg) from e

    if cfg:
        print("Stopping OrderlyWeb from '{}'".format(path))
        with docker_client() as client:
            if "proxy" in cfg.containers:
                stop_and_remove_container(client, cfg.containers["proxy"],
                                          kill)
            stop_and_remove_container(client, cfg.containers["web"], kill)
            stop_and_remove_container(client, cfg.containers["orderly"], kill)
            if network:
                remove_network(client, cfg.network)
            if volumes:
                for v in cfg.volumes.values():
                    remove_volume(client, v)
    else:
        print("OrderlyWeb not running from '{}'".format(path))



import io
from contextlib import redirect_stdout

import docker

import orderly_web
import orderly_web.cli
from orderly_web.docker_helpers import docker_client


def test_cli_basic_usage():
    path = "config/noproxy"
    orderly_web.cli.main(["start", path])
    st = orderly_web.status(path)
    assert st.is_running
    assert st.containers["web"]["status"] == "running"

    f = io.StringIO()
    with redirect_stdout(f):
        orderly_web.cli.main(["status", path])
    out = f.getvalue()
    assert str(st).strip() == out.strip()

    stop_args = ["stop", path, "--kill", "--volumes", "--network"]
    orderly_web.cli.main(stop_args)
    st = orderly_web.status(path)
    assert not st.is_running
    assert st.containers["web"]["status"] == "missing"



from orderly_web.docker_helpers import docker_client


def pull(cfg):
    print("Pulling orderly-web images:")
    with docker_client() as cl:
        for name, repo in cfg.images.items():
            image = str(repo)
            print("  - {} ({})".format(name, image))
            img = cl.images.pull(str(image))
            print("    `-> {}".format(img.short_id))



import io
from contextlib import redirect_stdout
import pytest
import shutil
import tempfile
import vault_dev
import yaml
import os

from orderly_web.config import *

sample_data = {"a": "value1", "b": {"x": "value2"}, "c": 1, "d": True,
               "e": None}


def test_config_string_reads_simple_values():
    assert config_string(sample_data, "a") == "value1"
    assert config_string(sample_data, ["a"]) == "value1"


def test_config_string_reads_nested_values():
    assert config_string(sample_data, ["b", "x"]) == "value2"


def test_config_string_throws_on_missing_keys():
    with pytest.raises(KeyError):
        config_string(sample_data, "x")
    with pytest.raises(KeyError):
        config_string(sample_data, ["b", "y"])


def test_config_none_is_missing():
    with pytest.raises(KeyError):
        config_string(sample_data, ["e"], False)
    assert config_string(sample_data, ["e"], True) is None


def test_config_string_validates_types():
    with pytest.raises(ValueError):
        config_string(sample_data, "c")


def test_config_string_default():
    assert config_string(sample_data, "x", True) is None


def test_config_integer():
    assert config_integer(sample_data, "c") == 1


def test_config_boolean():
    assert config_boolean(sample_data, "d")


def test_config_dict_returns_dict():
    assert config_dict(sample_data, ["b"]) == sample_data["b"]


def test_config_dict_strict_returns_dict():
    assert config_dict_strict(sample_data, ["b"], ["x"]) == sample_data["b"]


def test_config_dict_strict_raises_if_keys_missing():
    with pytest.raises(ValueError, match="Expected keys x, y for b"):
        config_dict_strict(sample_data, ["b"], ["x", "y"])


def test_config_dict_strict_raises_if_not_strings():
    dat = {"a": {"b": {"c": 1}}}
    with pytest.raises(ValueError, match="Expected a string for a:b:c"):
        config_dict_strict(dat, ["a", "b"], "c")


def test_config_enum_returns_string():
    assert config_enum(sample_data, ["b", "x"], ["value1", "value2"]) == \
        "value2"


def test_config_enum_raises_if_invalid():
    with pytest.raises(ValueError,
                       match=r"Expected one of \[enum1, enum2\] for b:x"):
        config_enum(sample_data, ["b", "x"], ["enum1", "enum2"])


def test_example_config():
    cfg = build_config("config/basic")
    assert cfg.network == "orderly_web_network"
    assert cfg.volumes["orderly"] == "orderly_web_volume"
    assert cfg.containers["orderly"] == "orderly_web_orderly"
    assert cfg.containers["web"] == "orderly_web_web"

    assert cfg.images["orderly"].repo == "vimc"
    assert cfg.images["orderly"].name == "orderly.server"
    assert cfg.images["orderly"].tag == "master"
    assert str(cfg.images["orderly"]) == "vimc/orderly.server:master"
    assert cfg.web_dev_mode
    assert cfg.web_port == 8888
    assert cfg.web_name == "OrderlyWeb"
    assert cfg.web_email == "admin@example.com"
    assert not cfg.web_auth_montagu
    assert cfg.web_auth_fine_grained
    assert cfg.web_auth_github_org == "vimc"
    assert cfg.web_auth_github_team == ""
    assert cfg.web_auth_github_app["id"] == "notarealid"
    assert cfg.web_auth_github_app["secret"] == "notarealsecret"
    assert cfg.sass_variables is None
    assert "css-generator" not in cfg.images
    assert "css" not in cfg.volumes
    assert cfg.logo_name is None
    assert cfg.logo_path is None
    assert cfg.favicon_path is None

    assert cfg.proxy_enabled
    assert cfg.proxy_ssl_self_signed
    assert str(cfg.images["proxy"]) == "vimc/orderly-web-proxy:master"

    assert cfg.orderly_initial_source == "clone"
    assert cfg.orderly_initial_url == \
        "https://github.com/reside-ic/orderly-example"

    assert cfg.slack_webhook_url == \
        "https://hooks.slack.com/services/T000/B000/XXXX"


def test_config_custom_styles():
    path = "config/customcss"
    cfg = build_config(path)
    expected_path = os.path.abspath(os.path.join(cfg.path, "variables.scss"))
    assert cfg.sass_variables == expected_path
    assert cfg.volumes["css"] == "orderly_web_css"
    assert cfg.images["css-generator"].repo == "vimc"
    assert cfg.images["css-generator"].name == "orderly-web-css-generator"
    assert cfg.images["css-generator"].tag == "master"
    assert cfg.logo_name == "my-test-logo.png"
    expected_path = os.path.abspath(os.path.join(cfg.path, "my-test-logo.png"))
    assert cfg.logo_path == expected_path
    expected_icon_path = os.path.abspath(
        os.path.join(cfg.path, "my-test-favicon.png"))
    assert cfg.favicon_path == expected_icon_path


def test_config_montagu():
    path = "config/montagu"
    cfg = build_config(path)
    assert cfg.montagu_url == "http://montagu"
    assert cfg.montagu_api_url == "http://montagu/api"


def test_string_representation():
    img = DockerImageReference("a", "b", "c")
    assert str(img) == "a/b:c"


def test_config_image_reference():
    data = {"foo": {
        "repo": "a", "name": "b", "tag": "c", "other": "d", "num": 1}}
    assert str(config_image_reference(data, "foo")) == "a/b:c"
    assert str(config_image_reference(data, ["foo"], "other")) == "a/d:c"
    with pytest.raises(KeyError):
        config_image_reference(data, ["foo"], "missing")
    with pytest.raises(ValueError):
        config_image_reference(data, ["foo"], "num")


def test_config_no_proxy():
    cfg = build_config("config/noproxy")
    assert not cfg.proxy_enabled


def test_config_proxy_not_enabled():
    options = {"proxy": {"enabled": False}}
    cfg = build_config("config/noproxy", options=options)
    assert not cfg.proxy_enabled


def test_can_substitute_secrets():
    with vault_dev.server() as s:
        cl = s.client()
        # Copy the certificates into the vault where we will later on
        # pull from from.
        cert = read_file("proxy/ssl/certificate.pem")
        key = read_file("proxy/ssl/key.pem")
        cl.write("secret/ssl/certificate", value=cert)
        cl.write("secret/ssl/key", value=key)
        cl.write("secret/db/password", value="s3cret")
        cl.write("secret/github/id", value="ghkey")
        cl.write("secret/github/secret", value="ghs3cret")
        cl.write("secret/ssh", public="public-key-data",
                 private="private-key-data")
        cl.write("secret/slack/webhook", value="http://webhook")

        # When reading the configuration we have to interpolate in the
        # correct values here for the vault connection
        cfg = build_config("config/complete")
        cfg.vault.url = "http://localhost:{}".format(s.port)
        cfg.vault.auth_args["token"] = s.token

        cfg.resolve_secrets()
        assert not cfg.proxy_ssl_self_signed
        assert cfg.proxy_ssl_certificate == cert
        assert cfg.proxy_ssl_key == key
        assert cfg.orderly_env["ORDERLY_DB_PASS"] == "s3cret"
        assert cfg.web_auth_github_app["id"] == "ghkey"
        assert cfg.web_auth_github_app["secret"] == "ghs3cret"
        assert cfg.orderly_ssh["private"] == "private-key-data"
        assert cfg.orderly_ssh["public"] == "public-key-data"


def test_combine():
    def do_combine(a, b):
        """lets us use combine with unnamed data"""
        combine(a, b)
        return a

    assert do_combine({"a": 1}, {"b": 2}) == \
        {"a": 1, "b": 2}
    assert do_combine({"a": {"x": 1}, "b": 2}, {"a": {"x": 3}}) == \
        {"a": {"x": 3}, "b": 2}
    assert do_combine({"a": {"x": 1, "y": 4}, "b": 2}, {"a": {"x": 3}}) == \
        {"a": {"x": 3, "y": 4}, "b": 2}
    assert do_combine({"a": None, "b": 2}, {"a": {"x": 3}}) == \
        {"a": {"x": 3}, "b": 2}


def combine_can_replace_dict():
    base = {"a": {"c": {"d": "x"}}, "b": "y"}
    options = {"a": {"c": None}}
    combine(base, options)
    assert base["a"]["c"] is None


def test_read_and_extra():
    with tempfile.TemporaryDirectory() as p:
        shutil.copy("config/basic/orderly-web.yml", p)
        with open("{}/patch.yml".format(p), "w+") as f:
            data = {"network": "patched_network"}
            yaml.dump(data, f)
        cfg = build_config(p, "patch")
        assert cfg.network == "patched_network"


def test_read_and_options():
    options = {"network": "patched_network"}
    cfg = build_config("config/basic", options=options)
    assert cfg.network == "patched_network"


def test_read_complex():
    with tempfile.TemporaryDirectory() as p:
        shutil.copy("config/basic/orderly-web.yml", p)
        data1 = {"network": "network1",
                 "volumes": {"proxy_logs": "mylogs"}}
        data2 = {"network": "network2",
                 "volumes": {"orderly": "mydata"}}
        with open("{}/patch.yml".format(p), "w+") as f:
            data = {"network": "patched_network"}
            yaml.dump(data1, f)
        cfg = build_config(p, "patch", data2)
        assert cfg.network == "network2"
        assert cfg.volumes["orderly"] == "mydata"
        assert cfg.volumes["proxy_logs"] == "mylogs"


def test_cant_overwrite_prefix_with_patch():
    with tempfile.TemporaryDirectory() as p:
        shutil.copy("config/basic/orderly-web.yml", p)
        with open("{}/patch.yml".format(p), "w+") as f:
            data = {"container_prefix": "patched_orderly_web"}
            yaml.dump(data, f)
        with pytest.raises(Exception,
                           match="'container_prefix' may not be modified"):
            build_config(p, "patch")


def test_cant_overwrite_prefix_with_options():
    with pytest.raises(Exception,
                       match="'container_prefix' may not be modified"):
        options = {"container_prefix": "patched_orderly_web"}
        build_config("config/basic", options=options)


def test_update_config_with_options_list():
    options = [{"network": "patched_network"}, {"web": {"dev_mode": False}}]
    data = build_config("config/basic", None, options=options)
    assert not data.web_dev_mode
    assert data.network == "patched_network"


def test_update_config_with_options_dict():
    options = {"network": "patched_network"}
    data = build_config("config/basic", None, options=options)
    assert data.web_dev_mode
    assert data.network == "patched_network"


def test_web_url_is_read_from_config():
    data = build_config("config/basic")
    assert data.web_url == "https://localhost"


def test_github_app_not_needed_if_using_montagu():
    options = {"web": {"auth": {"montagu": True,
                                "montagu_url": "whatever",
                                "montagu_api_url": "whatever"}}}
    data = build_config("config/basic", options=options)
    assert data.web_auth_github_app is None


def test_web_url_default_depends_on_proxy():
    options = {"web": {"url": None}}
    data = build_config("config/basic", options=options)
    assert data.web_url == "https://localhost:443"
    data = build_config("config/noproxy", options=options)
    assert data.web_url == "http://localhost:8888"


def test_web_url_required_if_not_proxied():
    with pytest.raises(Exception,
                       match="web_url must be provided"):
        options = {"web": {"url": None, "dev_mode": False}}
        build_config("config/noproxy", options=options)


def test_github_auth_required_if_not_using_montagu():
    with pytest.raises(KeyError, match="web:auth:github_org"):
        options = {"web": {"auth": {"github_org": None}}}
        cfg = build_config("config/basic", options=options)


def test_github_auth_ignored_if_using_montagu():
    options = {"web": {"auth": {"montagu": True,
                                "montagu_url": "https://localhost",
                                "montagu_api_url": "https://localhost"}}}
    cfg = build_config("config/basic", options=options)
    assert cfg.web_auth_github_app is None
    assert cfg.web_auth_github_org is None
    assert cfg.web_auth_github_team is None


def test_can_use_url_for_initial_source():
    url = "https://github.com/reside-ic/orderly-example"
    options = {"orderly": {"initial": {"source": "clone", "url": url}}}
    cfg = build_config("config/basic", options=options)
    assert cfg.orderly_initial_source == "clone"
    assert cfg.orderly_initial_url == url


def test_initial_clone_requires_url():
    options = {"orderly": {"initial": {"source": "clone"}}}
    with pytest.raises(KeyError, match="orderly:initial:url"):
        cfg = build_config("config/montagu", options=options)


def test_initial_demo_ignores_url():
    url = "https://github.com/reside-ic/orderly-example"
    options = {"orderly": {"initial": {"source": "demo", "url": url}}}
    f = io.StringIO()
    with redirect_stdout(f):
        cfg = build_config("config/basic", options=options)
    out = f.getvalue()
    assert "NOTE: Ignoring orderly:initial:url" in out


def read_file(path):
    with open(path, "r") as f:
        return f.read()



import docker

from orderly_web.config import read_config
from orderly_web.docker_helpers import docker_client


def print_status(path):
    print(status(path))


def status(path):
    return OrderlyWebStatus(path)


class OrderlyWebStatus:
    def __init__(self, path):
        self.path = path
        self.reload()

    def __str__(self):
        if self.cannot_read_status:
            return "Cannot read status from orderly-web because it has not " \
                "started successfully or is in an error state."
        if not self.is_running:
            return "<not running>"
        st_c = dict_map(self.containers, format_container)
        st_v = dict_map(self.volumes, format_volume)
        st_n = "Network: {}".format(self.network)
        ret = ["OrderlyWeb status:"]
        if st_c:
            ret += ["Containers:"] + st_c
        if st_v:
            ret += ["Volumes:"] + st_v
        ret += [st_n]
        return "\n".join(ret)

    def __repr__(self):
        return self.__str__()

    def reload(self):
        cfg_base = read_config(self.path)

        cfg_running = False

        try:
            cfg_running = cfg_base.fetch()
            self.cannot_read_status = False
        except docker.errors.NotFound:
            self.cannot_read_status = True

        self.is_running = bool(cfg_running)
        with docker_client() as client:
            self.containers = {k: container_status(client, v)
                               for k, v in cfg_base.containers.items()}
            if cfg_running:
                self.volumes = cfg_running.volumes
                self.network = cfg_running.network
            else:
                self.volumes = {}
                self.network = None


def format_container(role, status):
    return "  {}: {} ({})".format(role, status["status"], status["name"])


def format_volume(role, name):
    return "  {}: {}".format(role, name)


def container_status(client, name):
    try:
        status = client.containers.get(name).status
    except docker.errors.NotFound:
        status = "missing"
    return {"name": name, "status": status}


def dict_map(x, f):
    return [f(k, x[k]) for k in sorted(x.keys())]



import os
import pytest
import vault_dev

from orderly_web.vault import drop_envvar, resolve_secret, vault_config, \
    vault_not_enabled


def test_secret_reading():
    with vault_dev.server() as s:
        client = s.client()
        client.write("secret/foo", value="s3cret")
        assert resolve_secret("foo", client) == (False, "foo")
        assert resolve_secret("VAULT:secret/foo:value", client) == \
            (True, "s3cret")


def test_accessor_validation():
    with vault_dev.server() as s:
        client = s.client()
        with pytest.raises(Exception, match="Invalid vault accessor"):
            resolve_secret("VAULT:invalid", client)
        with pytest.raises(Exception, match="Invalid vault accessor"):
            resolve_secret("VAULT:invalid:a:b", client)


def test_error_for_missing_secret():
    with vault_dev.server() as s:
        client = s.client()
        msg = "Did not find secret at 'secret/foo'"
        with pytest.raises(Exception, match=msg):
            resolve_secret("VAULT:secret/foo:bar", client)


def test_error_for_missing_secret_key():
    with vault_dev.server() as s:
        client = s.client()
        client.write("secret/foo", value="s3cret")
        msg = "Did not find key 'bar' at secret path 'secret/foo'"
        with pytest.raises(Exception, match=msg):
            resolve_secret("VAULT:secret/foo:bar", client)


def test_vault_config():
    with vault_dev.server() as s:
        url = "http://localhost:{}".format(s.port)
        cfg = vault_config(url, "token", {"token": s.token})
        cl = cfg.client()
        assert cl.is_authenticated()


def test_vault_config_when_missing():
    cfg = vault_config(None, "token", {"token": "root"})
    cl = cfg.client()
    assert type(cl) == vault_not_enabled
    with pytest.raises(Exception, match="Vault access is not enabled"):
        cl.read("secret/foo")


# To run this test you will need a token for the vimc robot user -
# this can be found in the montagu vault as
# /secret/vimc-robot/vault-token
#
# This environment variable is configured on travis
def test_vault_config_login():
    with vault_dev.server() as s:
        cl = s.client()
        cl.sys.enable_auth_method(method_type="github")
        cl.write("auth/github/config", organization="vimc")

        url = "http://localhost:{}".format(s.port)
        token = os.environ["VAULT_TEST_GITHUB_PAT"]
        cfg = vault_config(url, "github", {"token": token})
        assert cfg.client().is_authenticated()


# Utility reuired to work around https://github.com/hvac/hvac/issues/421
def test_drop_envvar_removes_envvar():
    name = "VAULT_DEV_TEST_VAR"
    os.environ[name] = "x"
    drop_envvar(name)
    assert name not in os.environ


def test_drop_envvar_ignores_missing_envvar():
    name = "VAULT_DEV_TEST_VAR"
    drop_envvar(name)
    assert name not in os.environ



import requests
import json


class Notifier:
    def __init__(self, webhook):
        self.enabled = webhook is not None
        if self.enabled:
            self.url = webhook
            self.headers = {'Content-Type': 'application/json'}

    def post(self, message):
        if not self.enabled:
            return
        data = json.dumps({"text": message})
        # There are two ways that this can fail: (1) the slack server
        # can respond with an error (e.g., if the incoming webhook is
        # removed or expires, the slack server is down for
        # maintenence but still largely functioning, etc).  In that
        # case we'll get a nice HTTP error code and we can just print
        # that.  (2) the communication with the server fails because
        # the network is down, slack is really down, etc.  In that
        # case requests will Raise and that needs catching.
        #
        # In either case, as soon as one request fails, don't send
        # future notifications as they're highly unlikely to work,
        # and if they're timing out that'll get tedious.
        try:
            r = requests.post(self.url, data=data, headers=self.headers)
            if r.status_code >= 300:
                print("Problem sending message: " + r.reason)
                self.enabled = False
        except Exception as e:
            print("There was a problem sending the slack message:\n{}".format(
                str(e)))
            self.enabled = False



import base64

import pytest
from PIL import Image

from orderly_web.docker_helpers import *


def test_exec_safely_throws_on_failure():
    with docker_client() as cl:
        container = cl.containers.run("alpine", ["sleep", "10"],
                                      detach=True, auto_remove=True)
        with pytest.raises(Exception):
            exec_safely(container, "missing_command")
        container.kill()


def test_stop_and_remove_container_works():
    with docker_client() as cl:
        container = cl.containers.run("alpine", ["sleep", "10"],
                                      detach=True)
        name = container.name
        assert container_exists(cl, name)
        stop_and_remove_container(cl, name, False, 1)
        assert not container_exists(cl, name)


def test_string_into_container():
    with docker_client() as cl:
        container = cl.containers.run("alpine", ["sleep", "20"],
                                      detach=True, auto_remove=True)
        text = "a\nb\nc\n"
        string_into_container(text, container, "/test")
        out = container.exec_run(["cat", "/test"])
        assert out[0] == 0
        assert out[1].decode("UTF-8") == text
        container.kill()


def test_file_into_container():
    with docker_client() as cl:
        container = cl.containers.run("alpine", ["sleep", "20"],
                                      detach=True, auto_remove=True)
        img = Image.new("RGB", (60, 30), color="red")
        img.save("pil_red.png")
        file_into_container("pil_red.png", container, ".", "test_name.png")
        out = container.exec_run(["cat", "test_name.png"])
        with open("pil_red.png", "rb") as image:
            b64string = base64.b64encode(image.read())
        assert out[0] == 0
        assert base64.b64encode(out[1]) == b64string
        container.kill()
        os.remove("pil_red.png")


def test_container_wait_running_detects_start_failure():
    with docker_client() as cl:
        container = cl.containers.create("alpine")
        with pytest.raises(Exception) as e:
            container_wait_running(container, 0.1, 0.1)
        assert "is not running (created)" in str(e)


def test_container_wait_running_detects_slow_failure():
    with docker_client() as cl:
        with pytest.raises(Exception) as e:
            container = cl.containers.run("alpine", ["sleep", "1"],
                                          detach=True)
            container_wait_running(container, 0.1, 1.2)
        assert "was running but is now exited" in str(e)


def test_container_wait_running_returns_cintainer():
    with docker_client() as cl:
        container = cl.containers.run("alpine", ["sleep", "100"], detach=True)
        res = container_wait_running(container, 0.1, 1.2)
        assert res == container
        container.kill()
        container.remove()


def test_return_logs_and_remove_returns_stdout():
    with docker_client() as cl:
        result = return_logs_and_remove(cl, "alpine", ["echo", "1234"])
        assert "1234" in result


def test_return_logs_and_remove_returns_stderr():
    with docker_client() as cl:
        result = return_logs_and_remove(cl, "alpine", ["sh", "./nonsense"])
        assert "can't open './nonsense'" in result


def test_that_removing_missing_container_is_harmless():
    with docker_client() as cl:
        nm = "orderly_web_noncontainer"
        stop_and_remove_container(cl, nm, True)


def test_that_removing_missing_network_is_harmless():
    with docker_client() as cl:
        nm = "orderly_web_nonnetwork"
        remove_network(cl, nm)


def test_that_removing_missing_volume_is_harmless():
    with docker_client() as cl:
        nm = "orderly_web_nonvolume"
        remove_volume(cl, nm)



import math
import os
import tarfile
import tempfile
import time

import docker


def ensure_network(client, name):
    try:
        client.networks.get(name)
    except docker.errors.NotFound:
        print("Creating docker network '{}'".format(name))
        client.networks.create(name)


def ensure_volume(client, name):
    try:
        client.volumes.get(name)
    except docker.errors.NotFound:
        print("Creating docker volume '{}'".format(name))
        client.volumes.create(name)


def exec_safely(container, args):
    ans = container.exec_run(args)
    if ans[0] != 0:
        print(ans[1].decode("UTF-8"))
        raise Exception("Error running command (see above for log)")
    return ans


def return_logs_and_remove(client, image, args=None, mounts=None):
    try:
        container = client.containers.run(image, args, mounts=mounts,
                                          detach=True)
        res = container.wait()
        return container.logs().decode("UTF-8")
    finally:
        container.remove()


def stop_and_remove_container(client, name, kill, timeout=10):
    try:
        container = client.containers.get(name)
    except docker.errors.NotFound:
        return
    if container.status == "running":
        if kill:
            print("Killing '{}'".format(name))
            container.kill()
        else:
            print("Stopping '{}'".format(name))
            container.stop(timeout=timeout)
    print("Removing '{}'".format(name))
    container.remove()


def remove_network(client, name):
    try:
        nw = client.networks.get(name)
    except docker.errors.NotFound:
        return
    print("Removing network '{}'".format(name))
    nw.remove()


def remove_volume(client, name):
    try:
        v = client.volumes.get(name)
    except docker.errors.NotFound:
        return
    print("Removing volume '{}'".format(name))
    v.remove(name)


def container_exists(client, name):
    return docker_exists(client.containers, name)


def network_exists(client, name):
    return docker_exists(client.networks, name)


def volume_exists(client, name):
    return docker_exists(client.volumes, name)


def docker_exists(collection, name):
    try:
        collection.get(name)
        return True
    except docker.errors.NotFound:
        return False


# https://medium.com/@nagarwal/lifecycle-of-docker-container-d2da9f85959
def container_wait_running(container, poll=0.1, timeout=1):
    for i in range(math.ceil(timeout / poll)):
        if container.status != "created":
            break
        time.sleep(poll)
        container.reload()
    if container.status != "running":
        raise Exception("container '{}' ({}) is not running ({})".format(
            container.name, container.id[:8], container.status))
    time.sleep(timeout)
    container.reload()
    if container.status != "running":
        raise Exception("container '{}' ({}) was running but is now {}".format(
            container.name, container.id[:8], container.status))
    return container


def simple_tar(path, name):
    f = tempfile.NamedTemporaryFile()
    t = tarfile.open(mode="w", fileobj=f)
    abs_path = os.path.abspath(path)
    t.add(abs_path, arcname=name, recursive=False)
    t.close()
    f.seek(0)
    return f


def simple_tar_string(text, name):
    try:
        fd, tmp = tempfile.mkstemp(text=True)
        with os.fdopen(fd, "w") as f:
            f.write(text)
        return simple_tar(tmp, name)
    finally:
        os.remove(tmp)


# The python docker client does not provide nice 'docker cp' wrappers
# (https://github.com/docker/docker-py/issues/1771) - so we have to
# roll our own.  These are a real pain to do "properly".  For example
# see
# https://github.com/richfitz/stevedore/blob/845587/R/docker_client_support.R#L943-L1020
#
# So this function assumes that the destination directory exists and
# dumps out text into a file in the container
def string_into_container(txt, container, path):
    with simple_tar_string(txt, os.path.basename(path)) as tar:
        container.put_archive(os.path.dirname(path), tar)


def file_into_container(local_path, container, destination_path, name):
    tar = simple_tar(local_path, name)
    container.put_archive(destination_path, tar)


def string_from_container(container, path):
    stream, status = container.get_archive(path)
    try:
        fd, tmp = tempfile.mkstemp(text=False)
        with os.fdopen(fd, "wb") as f:
            for d in stream:
                f.write(d)
        with open(tmp, "rb") as f:
            t = tarfile.open(mode="r", fileobj=f)
            p = t.extractfile(os.path.basename(path))
            txt = p.readlines()
            return "\n".join([x.decode("utf8") for x in txt])
    finally:
        os.remove(tmp)


# There is an annoyance with docker and the requests library, where
# when the http handle is reclaimed a warning is printed.  It makes
# the test log almost impossible to read.
#
# https://github.com/kennethreitz/requests/issues/1882#issuecomment-52281285
# https://github.com/kennethreitz/requests/issues/3912
#
# This little helper can be used with python's with statement as
#
#      with docker_client() as cl:
#        cl.containers...
#
# and will close *most* unused handles on exit.  It's easier to look
# at than endless try/finally blocks everywhere.
class docker_client():
    def __enter__(self):
        self.client = docker.client.from_env()
        return self.client

    def __exit__(self, type, value, traceback):
        pass
